
c     The program "netCDF_climatology" performs the following actions:
c      *computes the monthly climatology for a specific variable
c      * the time span for statistical computations may be limited,
c         by specification of first and last years to consider
c      *either: appends the climatology to a non-empty netCDF file
c       or:     writes the climatology to an empty (existing)
c                 netCDF file
c     In any case, the climatological variables must be predefined on
c     the output netCDF file.

c     If no argument is given, the program terminates with an error
c     message which describes the correct syntax.

c     If the program terminates as intended, the file "ok_flag" is
c     touched in order to flag proper (seemingly) execution.


      program netCDF_climatology

      implicit none

      include 'netcdf.inc'
      include 'nc_dims.inc'
      include 'missing_values.inc'

c     Parameter (#entries required for computation of StD)
      integer min_defined
      parameter(min_defined = 2)

c     Variables/statistics:
      integer*2 out_data(nx,ny,12)
      integer n_data(nx,ny,12)
      real mean(nx,ny,12),std(nx,ny,12),
     >     minimum(nx,ny,12),maximum(nx,ny,12),first_value(nx,ny,12)
      real*8 sum(nx,ny,12),sum_squares(nx,ny,12)

c     Variables/others:
      integer*2 i2_field(nx,ny)
      integer   i4_field(nx,ny)
      real      field_in(nx,ny),
     >          lon_1d(nx),lat_1d(ny),lon_2d(nx,ny),lat_2d(nx,ny)
      real*8    r8_field(nx,ny)
      integer nc_status,org_length,origin_date(5),date(5),monthno,
     >        n_files,file_no,n_records,start(4),count(4),var_type,
     >        i,x,y,z,t,no,var_no,firstyear,lastyear,
     >        first_date(5),last_date(5),
     >        nc_id,var_id,month_id,time_id,t_id,clim_id,
     >        lon_id,lat_id,out_lon_id,out_lat_id,
     >        n_id,mean_id,std_id,min_id,max_id,
     >        climate_size,file_size,vrbl_size,size,
     >        n_dims,climate_dates_on,error
      real origin_sec,sec,first_sec,last_sec,timevalue,offset,scale,
     >     limit
      real*8 vrnce
      logical record_on,climate,got_date,
     >        n_ok,mean_ok,std_ok,min_ok,max_ok
      character*4 value
      character*10 vrbl_name
      character*20 units,date_out
      character*120 org_files(100)
      character*120 present_file,climate_file,unit_in


c     For command line arguments:
      integer iargc,narg

c     Functions:
      integer str_len,str_pos


      narg = iargc()

      if (narg .eq. 0) then
         write(*,*)
         write(*,*)'Syntax error.'
         write(*,*)'  Correct syntax:'
         write(*,*)'    nc_climatology /variable/ /firstyear/ ',
     >             '/lastyear </filelist/'
         write(*,*)'  where "nc_climatology" will add monthly"'
         write(*,*)'   climatology fields to a netCDF file.'
         write(*,*)'     /variable/  is the name of the variable under',
     >             ' consideration'
         write(*,*)'     /firstyear/ is the first year to consider ',
     >             '(optional)'
         write(*,*)'     /lastyear/  is the last year to consider'
         write(*,*)'                  (optional, if present, requires ',
     >             '/firstyear/)'
         write(*,*)'     /filelist/  is the name of a file that ',
     >             'contains'
         write(*,*)'                 * no. of file names on line 1'
         write(*,*)'                 * list of input file names on ',
     >             'following lines'
         write(*,*)'                 * output file name on the final ',
     >             'line'
         write(*,*)'                 (if there is only one input file,'
         write(*,*)'                  this file may also be the ',
     >             'output file)'
         write(*,*)'Terminates.'
         write(*,*)
         stop
      end if


c     Set 'limit', 'got_date':
      limit = 0.999*abs(miss_val_r(1))
      got_date = .false.


c     Initialize 'start', 'count':
      start(1) = 1
      start(2) = 1
      count(1) = nx
      count(2) = ny
      count(3) = 1
      count(4) = 1


c     Read variable name:
      call getarg(1,vrbl_name)
      vrbl_size = str_len(vrbl_name)

c     More input values?
      if (narg .ge. 2) then
         call getarg(2,value)
         read(value,*)firstyear
         if (narg .ge. 3) then
            call getarg(3,value)
            read(value,*)lastyear
         else
            lastyear = 9999
         end if
      else
         firstyear = -999
         lastyear = 9999
      end if

c     Read no. of files containing data,and the file to put the climatology:
      read(5,*)n_files
      if (n_files .lt. 1 .or. n_files .gt. 101) go to 999
      write(*,*)
      write(*,*)'... read no. of file names:',n_files
      write(*,*)


c     --- START main loop ---

c     Consider one level at a time:
      do z = 1,max(1,nz)


c        (Re-)Initialize arrays for statistics:
         do t = 1,12
            do y = 1,ny
               do x = 1,nx
                  n_data(x,y,t) = 0
                  sum(x,y,t) = dble(0.)
               end do
            end do
         end do


c        --- START file loop ---

         do file_no = 1,max(1,n_files-1)

c           Read the file name?
            if (z .eq. 1) read(5,'(a120)')org_files(file_no)
            present_file = org_files(file_no)
            file_size = str_len(present_file)

c           Open netCDF file:
            nc_status = NF_OPEN(present_file(1:file_size), NF_NOWRITE,
     >                          nc_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,1)

            write(*,*)'... opened file no.',file_no

c           Read lon/lat?
            if (n_files .gt. 1 .and. z .eq. 1 .and. file_no .eq. 1) then
               nc_status = NF_INQ_VARID(nc_id, 'lon', lon_id)
               if (nc_status .ne. NF_NOERR) call handle_err(nc_status,2)
               nc_status = NF_INQ_VARID(nc_id, 'lat', lat_id)
               if (nc_status .ne. NF_NOERR) call handle_err(nc_status,3)
               if (sphere_on .eq. 1) then
                  nc_status = NF_GET_VAR_REAL(nc_id, lon_id, lon_1d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,4)
                  nc_status = NF_GET_VAR_REAL(nc_id, lat_id, lat_1d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,5)
               else
                  nc_status = NF_GET_VAR_REAL(nc_id, lon_id, lon_2d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,6)
                  nc_status = NF_GET_VAR_REAL(nc_id, lat_id, lat_2d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,7)
               end if
            end if


c           Check horizontal dimensions, set variable's ID:
            record_on = .true.
            if (nz .eq. 0) then
               call check_nc_xy(nc_id,sphere_on,nx,ny,record_on,1,
     >                          vrbl_name,var_id,error)
               if (error .ne. 0) stop
               n_dims = 3
            else if (z .eq. 1) then
               call check_nc_xyz(nc_id,sphere_on,nx,ny,nz,record_on,1,
     >                           vrbl_name,var_id,error)
               if (error .ne. 0) stop
               n_dims = 4
            end if

c           Read time units, no. of records etc. from original file:
            nc_status = NF_INQ_VARID(nc_id, 'time', time_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,11)
            unit_in = ''
            nc_status = NF_GET_ATT_TEXT(nc_id, time_id, 'units',
     >                                  unit_in)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,12)
            nc_status = NF_GET_ATT_INT(nc_id, time_id,
     >                                 'climate_dates_on',
     >                                 climate_dates_on)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,13)
            nc_status = NF_INQ_DIMID(nc_id, 'time', t_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,14)
            nc_status = NF_INQ_DIMLEN(nc_id, t_id, n_records)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,15)

c           Set 'climate':
            if (climate_dates_on .eq. 1) then
               climate = .true.
            else
               climate = .false.
            end if

c           Find time origin, time units on input file:
            call string2date(unit_in,origin_date,origin_sec)
            size = str_pos(unit_in,1,' ') - 1
            units = unit_in(1:size)

c           Find the variable's type:
            nc_status = NF_INQ_VARTYPE(nc_id, var_id, var_type)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,32)

c           If type 'NF_SHORT' or 'NF_INT', set 'offset', 'scale':
            if (var_type .eq. NF_SHORT .or. var_type .eq. NF_INT) then
               nc_status = NF_GET_ATT_REAL(nc_id, var_id, 'add_offset',
     >                                     offset)
               if (nc_status .ne. NF_NOERR)
     >            call handle_err(nc_status,41)
               nc_status = NF_GET_ATT_REAL(nc_id, var_id,
     >                                     'scale_factor', scale)
               if (nc_status .ne. NF_NOERR)
     >            call handle_err(nc_status,42)
            end if


c           --- START time loop ---

            do no = 1, n_records


c              (Re-)Set 'start':
               if (n_dims .eq. 4) then
                  start(3) = z
                  start(4) = no
               else
                  start(3) = no
                  count(3) = 1
               end if

c              Read 'timevalue':
               nc_status = NF_GET_VAR1_REAL(nc_id, time_id, no,
     >                                      timevalue)
               if (nc_status .ne. NF_NOERR)
     >            call handle_err(nc_status,201)

c              Find the present date:
               call offset2date(climate,units,origin_date,
     >                          origin_sec,timevalue,date,sec)

               if (date(1) .ge. firstyear .and.
     >             date(1) .le. lastyear) then


                  write(*,101)'...processing depth no.',z,
     >                        ', file no.',file_no,', record no.',no
 101              format(6x,a,i2,a,i2,a,i4)

c                 --- START within requested range ---

c                 Update end dates:
                  call set_end_dates(date,sec,climate,got_date,
     >                               first_date,first_sec,
     >                               last_date,last_sec)

c                 Which month?
                  monthno = date(2)

c                 Read original field:
                  if (var_type .eq. NF_SHORT) then
                     nc_status = NF_GET_VARA_INT2(nc_id, var_id,
     >                                            start, count,
     >                                            i2_field)
                     if (nc_status .ne. NF_NOERR)
     >                  call handle_err(nc_status,211)
c                    Convert to real numbers:
                     call short2float_2D(nx,ny,i2_field,offset,scale,
     >                                   field_in)
                  else if (var_type .eq. NF_INT) then
                     nc_status = NF_GET_VARA_INT(nc_id, var_id,
     >                                           start, count, i4_field)
                     if (nc_status .ne. NF_NOERR)
     >                  call handle_err(nc_status,212)
c                    Convert to real numbers:
                     call int2float_2D(nx,ny,i4_field,offset,scale,
     >                                 field_in)
                  else if (var_type .eq. NF_FLOAT) then
                     nc_status = NF_GET_VARA_REAL(nc_id, var_id,
     >                                            start, count,
     >                                            field_in)
                     if (nc_status .ne. NF_NOERR)
     >                  call handle_err(nc_status,213)
                  else if (var_type .eq. NF_DOUBLE) then
                     nc_status = NF_GET_VARA_DOUBLE(nc_id, var_id,
     >                                              start, count,
     >                                              r8_field)
                     if (nc_status .ne. NF_NOERR)
     >                  call handle_err(nc_status,214)
                     write(*,*)'NOTE! Accuracy is lost from real*8 ',
     >                         'input.'
                     do y = 1,ny
                        do x = 1,nx
                           field_in(x,y) = sngl(r8_field(x,y))
                        end do
                     end do
                  else
                     go to 998
                  end if


c                 --- START processing statistics for this 2D field ---

c                 Update arrays for statistical computations etc.:
                  do y = 1,ny
                     do x = 1,nx
                        if (abs(field_in(x,y)) .lt. limit) then

c                          Look for new minimum and maximum values:
                           if (n_data(x,y,monthno) .eq. 0) then
                              minimum(x,y,monthno) = field_in(x,y)
                              maximum(x,y,monthno) = field_in(x,y)
                              first_value(x,y,monthno) = field_in(x,y)
                           else if (field_in(x,y) .lt.
     >                              minimum(x,y,monthno)) then
                              minimum(x,y,monthno) = field_in(x,y)
                           else if (field_in(x,y) .gt.
     >                              maximum(x,y,monthno)) then
                              maximum(x,y,monthno) = field_in(x,y)
                           end if

c                          Update 'n_data', 'sum', 'sum_squares':
                           n_data(x,y,monthno) = n_data(x,y,monthno) + 1
                           sum(x,y,monthno) = sum(x,y,monthno) +
     >                                        dble(field_in(x,y))
                           sum_squares(x,y,monthno) =
     >                             sum_squares(x,y,monthno) +
     >                             dble( (field_in(x,y) -
     >                                    first_value(x,y,monthno))*
     >                                   (field_in(x,y) -
     >                                    first_value(x,y,monthno)) )
                        end if
                     end do
                  end do

c                 --- END processing statistics for this 2D field ---


c                 --- END within requested range ---


               else

                  write(*,101)'...skipping depth no.',z,
     >                        ', file no.',file_no,', record no.',no

               end if

            end do

c           --- END time loop ---


            nc_status = NF_CLOSE(nc_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,301)

            write(*,*)'... closed file no.',file_no
            write(*,*)

         end do

c        --- END file loop ---


c        --- START compute statistics ---

         write(*,*)'... computing statistics:'

         do t = 1,12
            write(*,*)'   ... considering month no.',t
            do y = 1,ny
               do x = 1,nx

                  if (n_data(x,y,t) .eq. 0) then

                     mean(x,y,t) = miss_val_r(1)
                     std(x,y,t) = miss_val_r(1)
                     minimum(x,y,t) = miss_val_r(1)
                     maximum(x,y,t) = miss_val_r(1)

                  else
                     mean(x,y,t) = sngl( sum(x,y,t)/
     >                                    dble(n_data(x,y,t)) )
                     if (n_data(x,y,t) .ge. max(2,min_defined)) then
                        vrnce = ( sum_squares(x,y,t) -
     >                             dble(n_data(x,y,t))*
     >                              ( dble(mean(x,y,t)) -
     >                                dble(first_value(x,y,t)) )**2. )/
c     >                            ( dble(n_data(x,y,t)) - 1 )
     >                           ( dble(n_data(x,y,t)) )
                        if (vrnce .lt. dble(0.)) vrnce = dble(0.)
                        std(x,y,t) = sngl(dsqrt(vrnce))
                     else
                        std(x,y,t) = miss_val_r(1)
                     end if

                  end if

               end do
            end do
         end do
         write(*,*)

c        --- END compute statistics ---


         if (z .eq. 1) then
c           Set 'climate_file' to the name of the output file:
            if (n_files .gt. 1) then
               read(5,'(a120)')climate_file
            else
               climate_file = org_files(1)
            end if
            climate_size = str_len(climate_file)
         end if

c        (Re-)Open climatology file:
         nc_status = NF_OPEN(climate_file(1:climate_size), NF_WRITE,
     >                       clim_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,401)

         if (z .eq. 1) then

c           Get 'month' ID:
            nc_status = NF_INQ_VARID(clim_id, 'month', month_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,402)

c           Check that this file contains header information:
            call check_all_climate_dims(vrbl_name,clim_id,
     >                                  nx,ny,nz,sphere_on,
     >                                  n_id,n_ok,mean_id,mean_ok,
     >                                  std_id,std_ok,
     >                                  min_id,min_ok,max_id,max_ok)
            if (.not. n_ok .and. .not. mean_ok .and. .not. std_ok .and.
     >          .not. min_ok .and. .not. max_ok) go to 997

c           Output months:
            do t = 1,12
               nc_status = NF_PUT_VAR1_INT(clim_id, month_id, t, t)
               if (nc_status .ne. NF_NOERR)
     >            call handle_err(nc_status,403)
            end do

c           Output lon/lat?
            if (n_files .gt. 1) then
               nc_status = NF_INQ_VARID(clim_id, 'lon', out_lon_id)
               if (nc_status .ne. NF_NOERR)
     >            call handle_err(nc_status,411)
               nc_status = NF_INQ_VARID(clim_id, 'lat', out_lat_id)
               if (nc_status .ne. NF_NOERR)
     >            call handle_err(nc_status,412)
               if (sphere_on .eq. 1) then
                  nc_status = NF_PUT_VAR_REAL(clim_id, out_lon_id,
     >                                        lon_1d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,413)
                  nc_status = NF_PUT_VAR_REAL(clim_id, out_lat_id,
     >                                        lat_1d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,414)
               else
                  nc_status = NF_PUT_VAR_REAL(clim_id, out_lon_id,
     >                                        lon_2d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,415)
                  nc_status = NF_PUT_VAR_REAL(clim_id, out_lat_id,
     >                                        lat_2d)
                  if (nc_status .ne. NF_NOERR)
     >               call handle_err(nc_status,416)
               end if
            end if

         end if


c        --- START write results ---

c        Output no. of occurences:
         if (n_ok) then
            nc_status = NF_PUT_VAR_INT(clim_id, n_id, n_data)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,421)
         else
            write(*,*)'Could not find header information for "n_',
     >                vrbl_name(1:vrbl_size),'", skipping ...'
         end if


c        Output mean values:
         if (mean_ok) then
            nc_status = NF_GET_ATT_REAL(clim_id, mean_id, 'add_offset',
     >                                  offset)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,431)
            nc_status = NF_GET_ATT_REAL(clim_id, mean_id,
     >                                  'scale_factor', scale)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,432)
            call float2short_3D(nx,ny,12,mean,offset,scale,out_data)
            nc_status = NF_PUT_VAR_INT2(clim_id, mean_id, out_data)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,433)
         else
            write(*,*)'Could not find header information for "mean_',
     >                vrbl_name(1:vrbl_size),'", skipping ...'
         end if


c        Output standard deviations:
         if (std_ok) then
            nc_status = NF_GET_ATT_REAL(clim_id, std_id, 'add_offset',
     >                                  offset)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,441)
            nc_status = NF_GET_ATT_REAL(clim_id, std_id,
     >                                  'scale_factor', scale)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,442)
            call float2short_3D(nx,ny,12,std,offset,scale,out_data)
            nc_status = NF_PUT_VAR_INT2(clim_id, std_id, out_data)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,443)
         else
            write(*,*)'Could not find header information for "std_',
     >                vrbl_name(1:vrbl_size),'", skipping ...'
         end if

c        Output overall minimum values (by month):
         if (min_ok) then
            nc_status = NF_GET_ATT_REAL(clim_id, min_id, 'add_offset',
     >                                  offset)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,451)
            nc_status = NF_GET_ATT_REAL(clim_id, min_id,
     >                                  'scale_factor', scale)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,452)
            call float2short_3D(nx,ny,12,minimum,offset,scale,out_data)
            nc_status = NF_PUT_VAR_INT2(clim_id, min_id, out_data)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,453)
         else
            write(*,*)'Could not find header information for "min_',
     >                vrbl_name(1:vrbl_size),'", skipping ...'
         end if

c        Output overall maximum values (by month):
         if (max_ok) then
            nc_status = NF_GET_ATT_REAL(clim_id, max_id, 'add_offset',
     >                                  offset)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,461)
            nc_status = NF_GET_ATT_REAL(clim_id, max_id,
     >                                  'scale_factor', scale)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,462)
            call float2short_3D(nx,ny,12,maximum,offset,scale,out_data)
            nc_status = NF_PUT_VAR_INT2(clim_id, max_id, out_data)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,463)
         else
            write(*,*)'Could not find header information for "max_',
     >                vrbl_name(1:vrbl_size),'", skipping ...'
         end if

c        --- END write results ---


         write(*,*)' ... done computing, writing statistics for ',
     >                vrbl_name(1:vrbl_size),', level:',z

      end do

c     --- END main loop ---


c     Output first date:
      date_out = ''
      size = 20
      call date2string(size,0,first_date,first_sec,date_out)
      nc_status = NF_PUT_ATT_TEXT(clim_id, month_id, 'first_date',
     >                            size, date_out(1:size))
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,501)

c     Output last date:
      date_out = ''
      size = 20
      call date2string(size,0,last_date,last_sec,date_out)
      nc_status = NF_PUT_ATT_TEXT(clim_id, month_id, 'last_date',
     >                            size, date_out(1:size))
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,502)


c     Close file w/climatology:
      nc_status = NF_CLOSE(clim_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,991)

      write(*,*)
      write(*,*)'The program has apparently finished successfully.'
      write(*,*)

      open(41, file= 'ok_flag', status='unknown')
      close(41)

      go to 1000

 997  write(*,*)
      write(*,*)'Error in the program "nc_climatology".'
      write(*,*)' No climatology header exists for the variable: ',
     >          vrbl_name(1:vrbl_size)
      write(*,*)'Terminates.'
      write(*,*)
      nc_status = NF_CLOSE(clim_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,997)
      stop

 998  write(*,*)
      write(*,*)'Error in the program "nc_climatology".'
      write(*,*)' Unrecognized type of variable:',var_type
      write(*,*)' Recognized:'
      write(*,*)'  NF_SHORT:  ',NF_SHORT
      write(*,*)'  NF_INT:    ',NF_INT
      write(*,*)'  NF_FLOAT:  ',NF_FLOAT
      write(*,*)'  NF_DOUBLE: ',NF_DOUBLE
      write(*,*)'Terminates.'
      write(*,*)
      nc_status = NF_CLOSE(nc_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,998)
      stop

 999  write(*,*)
      write(*,*)'Error in the program "nc_climatology".'
      write(*,*)' Illegal no. of files specified, must be >0, <101.'
      write(*,*)' Specified value:',n_files
      write(*,*)'Terminates.'
      write(*,*)

 1000 end

