
ccc   Subroutines on "check_nc_dims.f":
ccc    check_all_climate_dims
ccc                     - checks existence and dimensional sizes of
ccc                       climatological variables for a specific
ccc                       "base" variable
ccc    check_one_climate_dims
ccc                     - checks existence and dimensional sizes of a
ccc                       specific (climatological) variable
ccc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


c     ********************
c     General description:
c     ********************

c     The subroutine "check_all_climate_dims" should be used to check
c     the existence and dimensions of the climatological variables
c     (assocoated with the "base variable" 'vrbl_name') for the list
c     on the file 'stats_texts.inc'. Note that additions on this
c     include file will have no effect unless this subroutine is
c     modified accordingly. This subroutine returns (non-)existence
c     flags and IDs for the climatoligical variables with headers on
c     'clim_id'.


c     ***********************
c     Input/output variables:
c     ***********************

c      vrbl_name      (s)  - name of the variable for which climatology
c                             is requested to be added
c      clim_id        (i)  - netCDF (climatology) file ID
c      nx, ny, nz     (i)  - grids should be nx by ny by nz; nz may be 0
c      nx, ny, nz     (i)  - grids should be nx by ny by nz
c      sphere_on      (i)  - 1: horizontal grid should be lon,lat
c                            0: horizontal grid should be x,y
c      n_id, mean_id, std_id,
c       min_id,max_id (i)  - the output variable ID for the various
c                             climatology variables associated with
c                             'vrbl_name'
c      n_ok, mean_ok, std_ok,
c       min_ok,max_ok (l)  - the  existence (.true.) /
c                              nonexistence (.false.) flags for the
c                             various climatology variables

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      l  - logical          [ (.false.)/0 or (.true.)/nonzero ]
c      s  - string           [ character*(*) ]


      subroutine check_all_climate_dims(vrbl_name,clim_id,
     >                                  nx,ny,nz,sphere_on,
     >                                  n_id,n_ok,mean_id,mean_ok,
     >                                  std_id,std_ok,
     >                                  min_id,min_ok,max_id,max_ok)

      implicit none

      include 'stats_texts.inc'

c     Input:
      integer clim_id,nx,ny,nz,sphere_on
      character*(*) vrbl_name

c     Output:
      integer n_id,mean_id,std_id,min_id,max_id
      logical n_ok,mean_ok,std_ok,min_ok,max_ok

c     Local:
      integer vrbl_size,clim_size
      character*15 clim_name

c     Function:
      integer str_len


      vrbl_size = str_len(vrbl_name)


c     Consider no. of entries:
      clim_size = str_len(n_name)
      clim_name = n_name(1:clim_size) // vrbl_name(1:vrbl_size)
      call check_one_climate_dims(clim_id,clim_name,nx,ny,nz,sphere_on,
     >                            n_id,n_ok)


c     Consider mean values:
      clim_size = str_len(mean_name)
      clim_name = mean_name(1:clim_size) // vrbl_name(1:vrbl_size)
      call check_one_climate_dims(clim_id,clim_name,nx,ny,nz,sphere_on,
     >                            mean_id,mean_ok)


c     Consider standard deviations:
      clim_size = str_len(std_name)
      clim_name = std_name(1:clim_size) // vrbl_name(1:vrbl_size)
      call check_one_climate_dims(clim_id,clim_name,nx,ny,nz,sphere_on,
     >                            std_id,std_ok)


c     Consider minimum values:
      clim_size = str_len(min_name)
      clim_name = min_name(1:clim_size) // vrbl_name(1:vrbl_size)
      call check_one_climate_dims(clim_id,clim_name,nx,ny,nz,sphere_on,
     >                            min_id,min_ok)


c     Consider maximum values:
      clim_size = str_len(max_name)
      clim_name = max_name(1:clim_size) // vrbl_name(1:vrbl_size)
      call check_one_climate_dims(clim_id,clim_name,nx,ny,nz,sphere_on,
     >                            max_id,max_ok)

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "check_one_climate_dims" should be used to check
c     the existence and x,y(,z)-dimensions of a particular
c     (climatological) variable. The (non-)existence is flagged, and
c     if the requested variable is found, the dimensions are checked.
c     If these are OK, the subroutine returns and passes the
c     (climatological) variable's ID.


c     ***********************
c     Input/output variables:
c     ***********************

c      clim_id        (i)  - netCDF (climatology) file ID
c      clim_name      (s)  - name of the climatology variable
c      nx, ny, nz     (i)  - grids should be nx by ny by nz; nz may be 0
c      sphere_on      (i)  - 1: horizontal grid should be lon,lat
c                            0: horizontal grid should be x,y
c      nx, ny, nz     (i)  - grids should be nx by ny by nz
c      sphere_on      (i)  - 1: horizontal grid should be lon,lat
c                            0: horizontal grid should be x,y
c      clim_var_id    (i)  - the output variable ID for 'clim_name' on
c                             the file associated with 'clim_id'
c      clim_var_ok    (l)  - flag for existence (.true.) /
c                                  nonexistence (.false.)
c                             of 'clim_name' on 'clim_id'


      subroutine check_one_climate_dims(clim_id,clim_name,
     >                                  nx,ny,nz,sphere_on,
     >                                  clim_var_id,clim_var_ok)

      implicit none

c     Input:
      integer clim_id,nx,ny,nz,sphere_on
      character*(*) clim_name

c     Output:
      integer clim_var_id
      logical clim_var_ok

c     Local:
      integer error
      logical record_on


c     Check dimaensions on climatology file:
      record_on = .false.
      if (nz .eq. 0) then
         call check_nc_xy(clim_id,sphere_on,nx,ny,record_on,0,
     >                     clim_name,clim_var_id,error)
      else
         call check_nc_xyz(clim_id,sphere_on,nx,ny,nz,record_on,0,
     >                     clim_name,clim_var_id,error)
      end if

      if (error .eq. -1) then
         clim_var_ok = .false.
         go to 1000
      else if (error .gt. 0) then
         go to 999
      else
         clim_var_ok = .true.
      end if


      go to 1000

 999  write(*,*)
      write(*,*)'Nonacceptable error code returned from "check_nc_*":',
     >          error
      write(*,*)'Terminates in "check_one_climate_dims".'
      write(*,*)
      stop

 1000 return
      end

