
c     Subroutine on this file:
c      add_nc_month_header     - adds header for the variable 'month'
c      add_nc_var_climate      - framework for adding headers for
c                                 climatological quantities
c      add_nc_climate_header   - adds a header for one climatological
c                                 field

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


c     ********************
c     General description:
c     *********************

c     The subroutine "add_nc_month_header" adds a header for the
c      climatological time variable 'month', and its attributes,
c      provided these don't already exist.


      subroutine add_nc_month_header()

      implicit none

      include 'netcdf.inc'
      include 'id.inc'
      include 'err_codes.inc'

c     Local:
      integer nc_status


c     Explore existence:
      nc_status = NF_INQ_VARID(nc_id, 'month', month_id)

c     Define?
      if (nc_status .eq. no_var_match) then
c        Define variable:
         nc_status = NF_DEF_VAR(nc_id, 'month', NF_INT, 1, m_id,
     >                          month_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,601)
      else if (nc_status .eq. NF_NOERR) then
         write(*,*)'  ... variable "month" previously ',
     >             'defined. Continuing execution.'
      else
         call handle_err(nc_status,602)
      end if

      call nc_char_attrib(month_id,'long_name','Climatological months',
     >                    'month')
      call nc_char_attrib(month_id,'units',
     >                    'month no. (1:January ... 12:December)',
     >                    'month')
      call nc_char_attrib(month_id,'method','average','month')
      call nc_char_attrib(month_id,'first_date',
     >                    'xxxx-xx-xx xx:xx:xx.x','month')
      call nc_char_attrib(month_id,'last_date',
     >                    'xxxx-xx-xx xx:xx:xx.x','month')
      write(*,*)' Included variable   month    / Climatological months'

      return
      end



c     ********************
c     General description:
c     *********************

c     The subroutine "add_nc_var_climate" is a framework for adding
c      headers for climatological quantities related to netCDF
c      time-dependent fields. The headers for one field is created
c      by calling "add_nc_climate_header".


c     ****************
c     Input variables:
c     ****************

c      var_id         (i)  - the netCDF variable for which headers of
c                             associated climatological fields will be
c                             added
c      units          (s)  - the units of the variable associated with
c                             'var_id'
c      entry_no       (i)  - the position of the present variable in
c                             the list on the include file "name.inc'
c      offset         (r)  - offset used in 'field' upon output
c      scale          (r)  - scale factor used in 'field' upon output
c      dim_options    (i)  - no. of sets of different (x,y,z,t)-dimensions
c      n_dim          (iA) - no. of dimensions for the various
c                             'dim_options' dimensional combinations
c      dim_ids        (iA) - netCDF IDs for the various dimensional
c                             combinations

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      s  - string           [ character*(*) ]
c      iA - integer array    [ 4 byte, aka. "long" ]


      subroutine add_nc_var_climate(var_id,units,entry_no,offset,scale,
     >                              dim_options,n_dims,dim_ids)

      implicit none

      include 'netcdf.inc'
      include 'id.inc'
      include 'name.inc'
      include 'stats_texts.inc'

c     Input:
      integer var_id,entry_no,
     >        dim_options,n_dims(dim_options),dim_ids(4,dim_options)
      real scale,offset
      character*(*) units

c     Local:
      integer n_id,mean_id,std_id,min_id,max_id,sq_in,ref_id,sq_id,
     >        dim_option,clim_dims,clim_dim_ids(4),n
      character*10 vrbl_name


c     Set dimensional IDs for the climatological variables:
      dim_option = var_dimension(entry_no)
      clim_dims = n_dims(dim_option)
      do n = 1,clim_dims-1
         clim_dim_ids(n) = dim_ids(n,dim_option)
      end do
      clim_dim_ids(clim_dims) = m_id
      vrbl_name = var_name(entry_no)

c     no. of occurences:
      call add_nc_climate_header(vrbl_name,clim_dims,clim_dim_ids,
     >                           offset,scale,n_name,n_text,'none',n_id)

c     mean values:
      call add_nc_climate_header(vrbl_name,clim_dims,clim_dim_ids,
     >                           offset,scale,mean_name,mean_text,
     >                           units,mean_id)

c     standard deviations:
      call add_nc_climate_header(vrbl_name,clim_dims,clim_dim_ids,
     >                           offset,scale,std_name,std_text,
     >                           units,std_id)

c     overall minimum values (by month):
      call add_nc_climate_header(vrbl_name,clim_dims,clim_dim_ids,
     >                           offset,scale,min_name,min_text,
     >                           units,min_id)

c     overall maximum values (by month):
      call add_nc_climate_header(vrbl_name,clim_dims,clim_dim_ids,
     >                           offset,scale,max_name,max_text,
     >                           units,min_id)

      return
      end



c     ********************
c     General description:
c     *********************

c     The subroutine "add_nc_climate_header" includes
c      * a new climatological variable on the file with netCDF ID 'nc_id'
c      * a name of the new variable       (from 'short_text' and 'var_name')
c      * a  string  attribute "long_name" (from 'long_text' and 'var_name')
c      * a  string  attribute "units"
c      * an         attribute "missing_value"


c     ***********************
c     Input/output variables:
c     ***********************

c      var_name       (s)  - name of the existing variable for which
c                             climatological information will be added
c      n_dims         (i)  - no. of dimensions
c                             for the climatological variables
c      out_dim_ids    (iA) - netCDF IDs for the 'n_dims' dimensions
c      offset         (r)  - offset used in 'field' upon output
c      scale          (r)  - scale factor used in 'field' upon output
c      short_text     (s)  - head of (climatology) variable name
c      long_text      (s)  - head of (climatology) attribute 'long_name'
c      units          (s)  - units for the present (climatology) variable,
c                             often the same as the units for 'var_name'
c      tmp_id         (i)  - output, netCDF (climatology) variable ID

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      s  - string           [ character*(*) ]


      subroutine add_nc_climate_header(var_name,n_dims,out_dim_ids,
     >                                 offset,scale,
     >                                 short_text,long_text,
     >                                 units,tmp_id)

      implicit none

      include 'netcdf.inc'
      include 'id.inc'
      include 'err_codes.inc'

c     Input:
      integer n_dims,out_dim_ids(4),var_type
      real scale,offset
      character*(*) var_name
      character*(*) short_text
      character*(*) long_text
      character*(*) units
c     Output:
      integer tmp_id

c     Local:
      integer size,name_size,unit_size,short_size,long_size,
     >        attributes_option,nc_status
      character*80 name,longname

c     Function:
      integer str_len


c     Set string sizes:
      size = str_len(var_name)
      short_size = str_len(short_text)
      long_size = str_len(long_text)


c     Set name of variable:
      name = short_text(1:short_size) // var_name(1:size)
      name_size = str_len(name)


c     Explore existence:
      nc_status = NF_INQ_VARID(nc_id, name(1:name_size), tmp_id)

c     Define?
      if (nc_status .eq. no_var_match) then
c        Define variable:
         if (short_text(1:2) .eq. 'n_') then
            nc_status = NF_DEF_VAR(nc_id, name(1:name_size), NF_INT,
     >                             n_dims, out_dim_ids, tmp_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,151)
         else
            nc_status = NF_DEF_VAR(nc_id, name(1:name_size), NF_SHORT,
     >                             n_dims, out_dim_ids, tmp_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status,152)
         end if
      else if (nc_status .eq. NF_NOERR) then
         write(*,*)'  ... variable "',name(1:name_size),'" previously ',
     >             'defined. Continuing execution.'
      else
         call handle_err(nc_status,152)
      end if


c     Variable attributes 'long_name', 'units':
      longname = long_text(1:long_size) // ' '
      longname = longname(1:long_size+1) // var_name
      call nc_char_attrib(tmp_id,'long_name',longname,name)
      call nc_char_attrib(tmp_id,'units',units,name)

c     If variable is real (i.e., not no. of occurences),
c     variable attributes 'scale_factor', 'add_offset':
      if (short_text(1:short_size) .ne. 'n_') then
         call nc_float_attrib(tmp_id,'scale_factor',1,scale,name)
         call nc_float_attrib(tmp_id,'add_offset',1,offset,name)
      end if

c     Variable attribute 'missing_value':
      call nc_miss_val_i2(tmp_id, name)


      return
      end
