
c     Subroutine on this file:
c      write_all_nc_headers   - main subroutine for creating
c                                netCDF headers

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


c     - NOTES FOR "write_all_nc_headers" -

c     **************
c     General notes:
c     **************

c     1. The path of the include file ('netdcf.inc') must be
c        included when compiling!; e.g.,
c          f77 -c -I/usr/local/netcdf-3.4/include write_all_nc_headers.f

c     2. "write_all_nc_headers" requires that the file which will
c        contain the header (see the file "specs.inc") has not been
c        opened outside "write_all_nc_headers".

c     3. "write_all_nc_headers" does not close the netCDF-file,
c        but exits in "data mode".


c     *********************************************
c     List of implemented parameters and variables:
c     *********************************************

c      (*1), (*2), (*3) : See comments below

c      parameter (short/long) type  dimensions  accuracy units      id
c      ^^^^^^^^^^^^^^^^^^^^^^ ^^^^  ^^^^^^^^^^  ^^^^^^^^ ^^^^^      ^^

c     COMPULSORY:
c      lon / Longitude        float (*1)        N/A      degrees    lon_id
c      lat / Latitude         float (*1)        N/A      degrees    lat_id
c      time / Time            float nt          N/A      (*1)       time_id

c     ... COMPULSORY IF VERTICAL RESOLUTION:
c      level / (*2)           int4  nz          (*3)     (*3)       level_id

c     ... COMPULSORY IF ICE:
c      ice_level / Ice levels int4  nz_ice      N/A      layer no.  ice_level_id

c     ANY OF:
c      temp / Temperature     int2  nx,ny,nz,nt 0.01     deg.
c                                                        Kelvin     temp_id
c      salt / Salinity        int2  nx,ny,nz,nt 0.01     PSU        salt_id
c      sst / Sea surface
c            temperature      int2  nx,ny,nt    0.01     deg.
c                                                        Kelvin     sst_id
c      sss / Sea surface
c            salinity         int2  nx,ny,nt    0.01     PSU        sss_id
c      sgm_t / Sigma-t
c              (density)      int2  nx,ny,nz,nt 0.01     kg/(m**3)  sgm_t_id
c      thck  / Layer thickness
c                             int2  nx,ny,nz,nt 0.1      m          thck_id
c      u / Velocity-x         int2  nx,ny,nz,nt 0.001    m/s        u_id
c      v / Velocity-y         int2  nx,ny,nz,nt 0.001    m/s        v_id
c      vol_x / Volume flux-x
c             (layers)        int2  nx,ny,nz,nt 0.1      (m**2)/s   vol_x_id
c      vol_y / Volume flux-y
c             (layers)        int2  nx,ny,nz,nt 0.1      (m**2)/s   vol_y_id
c      ice_frac / Ice cover
c                 (fraction)  int2  nx,ny,nt    0.001    none       ice_frac_id
c      ice_thck / Ice
c                 thickness   int2  nx,ny,
c                                   nz_ice,nt   0.001    m          ice_thck_id
c      u_ice / Ice velocity-x int2  nx,ny,nt    0.001    m/s        u_ice_id
c      v_ice / Ice velocity-y int2  nx,ny,nt    0.001    m/s        v_ice_id
c      vol_x_ice / Ice volume
c                  flux-x
c                  (layers)   int2  nx,ny,nt    0.001    (m**2)/s   vol_x_ice_id
c      vol_y_ice / Ice volume
c                  flux-y
c                  (layers)   int2  nx,ny,nt    0.001    (m**2)/s   vol_y_ice_id
c      ssh / Sea surface
c            height           int2  nx,ny,nt    0.001    m          ssh_id
c      topo / Bottom
c             topography      float nx,ny       N/A      m          topo_id


c     (*1):
c     'time' is set relative to 'reference_date', with units as specified
c     by 'timeunit'. These are both set in "specs.inc", and recognized
c     values of 'timeunit' are:
c      'seconds'
c      'minutes'
c      'hours'
c      'days'
c      'months'
c      'years'

c     (*2):
c     'lon'/'lat' dimensions depend on the grid ('sphere_on' in "specs.inc"):
c      parameter
c       (short/long name)   'sphere_on'  dimensions
c      ^^^^^^^^^^^^^^^^^^   ^^^^^^^^^^^  ^^^^^^^^^^
c      lon / Longitude      0            nx,ny
c      lon / Longitude      1            nx
c      lat / Latitude       0            nx,ny
c      lat / Latitude       1            ny

c     (*3):
c     The possible types of vertical levels ('leveltype' in "specs.inc") are:
c      type levels            accuracy  units
c      ^^^^ ^^^^^^            ^^^^^^^^  ^^^^^
c      1    Layers            N/A       none
c      2    z (positive)      1.0       m
c      3    z (negative)      1.0       m
c      4    sigma (positive)  0.001     none
c      5    sigma (negative)  0.001     none

c     Misc.:
c      int2: 2 byte integer (aka. "short")
c      int4: 4 byte integer (aka. "long")


      subroutine write_all_nc_headers()


      implicit none

      include 'netcdf.inc'
      include 'id.inc'
      include 'specs.inc'

c     Local:
      integer dim_options
      parameter (dim_options = 9)
      integer nc_status,size,filesize,var_n_dim(dim_options),
     >        var_dim_ids(4,dim_options),option_no
      real range(2)
      logical exists
      character reply
      integer i

c     Function:
      integer str_len


      filesize = str_len(filename)


c     ************
c     netCDF Open:
c     ************

c     Open file, check existence:
      nc_status = NF_CREATE(filename, NF_NOCLOBBER, nc_id)
      if (nc_status .ne. NF_NOERR) then
         write(*,*)
         write(*,*)'The file you specified:'
         write(*,*)'    ',filename(1:filesize)
         write(*,*)' probably exists!'
         write(*,*)'Continue to add additional header information',
     >             ' to this file? (y/n)'
         read(*,'(a1)')reply
         do while (reply .ne. 'y' .and. reply .ne. 'n')
            write(*,*)'Please reply by typing one of',
     >                " the characters 'y' or 'n':"
            read(*,'(a1)')reply
         end do
         if (reply .eq. 'n') then
            write(*,*)
            write(*,*)'Terminates as requested.'
            write(*,*)
            stop
         else
c           Create file failed, retry opening file, put in define mode:
            nc_status = NF_OPEN(filename, NF_WRITE, nc_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status, 1)
            nc_status = NF_REDEF(nc_id)
            if (nc_status .ne. NF_NOERR) call handle_err(nc_status, 2)
         end if

         exists = .true.

      else

         exists = .false.

      end if

      write(*,*)
      write(*,*)'================================================='
      if (.not. exists) then
         write(*,*)' Attempting to create header for the netCDF file'
      else
         write(*,*)' Attempting to modify header for the netCDF file'
      end if
      write(*,*)'     ',filename(1:filesize)
      write(*,*)'================================================='
      write(*,*)



c     *************************
c     netCDF global attributes:
c     *************************

c     General information:
c     --------------------

      call nc_char_attrib(NF_GLOBAL,'title',title,'global')
      call nc_char_attrib(NF_GLOBAL,'source',source,'global')
      call nc_char_attrib(NF_GLOBAL,'method',method,'global')
      call nc_char_attrib(NF_GLOBAL,'reference',reference,'global')
      call nc_char_attrib(NF_GLOBAL,'web_reference',web_reference,
     >                    'global')
      call nc_char_attrib(NF_GLOBAL,'history',history,'global')
      call nc_char_attrib(NF_GLOBAL,'author',author,'global')
      call nc_char_attrib(NF_GLOBAL,'email',email,'global')


c     ******************
c     netCDF dimensions:
c     ******************

      write(*,*)
      write(*,*)'Dimensions:'

      if (sphere_on .eq. 1) then
         call write_nc_dim_header('lon',nx,x_id)
         call write_nc_dim_header('lat',ny,y_id)
      else
         call write_nc_dim_header('x',nx,x_id)
         call write_nc_dim_header('y',ny,y_id)
      end if
      if (nz .gt. 0) call write_nc_dim_header('z',nz,z_id)
      if (nz_ice .gt. 0)
     >   call write_nc_dim_header('z_ice',nz_ice,z_ice_id)
      call write_nc_dim_header('time',NF_UNLIMITED,t_id)
      if (climatology_on .eq. 1)
     >   call write_nc_dim_header('month',12,m_id)


c     --- START set arrays with dimension id-s ---

c     option 1: x (longitude)
      var_n_dim(1) = 1
      var_dim_ids(1,1) = x_id

c     option 2: y (latitude)
      var_n_dim(2) = 1
      var_dim_ids(1,2) = y_id

c     option 3: z
      var_n_dim(3) = 1
      var_dim_ids(1,3) = z_id

c     option 4: z_ice
      var_n_dim(4) = 1
      var_dim_ids(1,4) = z_ice_id

c     option 5: t
      var_n_dim(5) = 1
      var_dim_ids(1,5) = t_id

c     option 6: x,y
      var_n_dim(6) = 2
      var_dim_ids(1,6) = x_id
      var_dim_ids(2,6) = y_id

c     option 7: x,y,t
      var_n_dim(7) = 3
      var_dim_ids(1,7) = x_id
      var_dim_ids(2,7) = y_id
      var_dim_ids(3,7) = t_id

c     option 8: x,y,z,t
      var_n_dim(8) = 4
      var_dim_ids(1,8) = x_id
      var_dim_ids(2,8) = y_id
      var_dim_ids(3,8) = z_id
      var_dim_ids(4,8) = t_id

c     option 9: x,y,z,t
      var_n_dim(9) = 4
      var_dim_ids(1,9) = x_id
      var_dim_ids(2,9) = y_id
      var_dim_ids(3,9) = z_ice_id
      var_dim_ids(4,9) = t_id

c     --- END set arrays with dimension id-s ---


c     *****************
c        ***********

c     *****************
c     netCDF Variables:
c     *****************


c     Grid specification:
c     %%%%%%%%%%%%%%%%%%%


      write(*,*)
      write(*,*)'Grid specification:'


c     Longitude:
      call write_nc_var_header
     >        ('lon',NF_FLOAT,dim_options,var_n_dim,var_dim_ids,
     >         lon_id,1,0.,0.)
      range(1) = -360.
      range(2) = 720.
      call nc_float_attrib(lon_id,'valid_range',2,range,'lon')

c     Latitude:
      call write_nc_var_header
     >        ('lat',NF_FLOAT,dim_options,var_n_dim,var_dim_ids,
     >         lat_id,1,0.,0.)
      range(1) = -90.
      range(2) = 90.
      call nc_float_attrib(lat_id,'valid_range',2,range,'lat')

      if (nz .gt. 0)
c        Levels:
     >      call write_nc_var_header
     >              ('level',NF_INT,dim_options,var_n_dim,var_dim_ids,
     >               level_id,1,level_offset,level_scale)

      if (nz_ice .gt. 0)
c        Ice levels:
     >      call write_nc_var_header
     >              ('ice_level',NF_INT,
     >               dim_options,var_n_dim,var_dim_ids,
     >               ice_level_id,1,ice_level_offset,ice_level_scale)

c     Time:
      call write_nc_time_header(dim_options,var_n_dim,var_dim_ids)

      if (climatology_on .eq. 1)
c        Month:
     >   call add_nc_month_header()


c     Hydrography:
c     %%%%%%%%%%%%

      write(*,*)
      write(*,*)'Hydrography:'

      if (nz .gt. 0) then

c        Temperature:
         if (temp_on .eq. 1)
     >      call write_nc_var_header
     >              ('temp',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               temp_id,1,temp_offset,temp_scale)

c        Temperature (Kelvin):
         if (temp_K_on .eq. 1)
     >      call write_nc_var_header
     >              ('temp_K',NF_SHORT,
     >               dim_options,var_n_dim,var_dim_ids,
     >               temp_K_id,1,temp_K_offset,temp_K_scale)

c        Salinity:
         if (salt_on .eq. 1)
     >      call write_nc_var_header
     >              ('salt',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               salt_id,salt_offset,salt_scale)

c        Sigma-t (density):
         if (sgm_t_on .eq. 1)
     >      call write_nc_var_header
     >              ('sgm_t',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               sgm_t_id,1,sgm_t_offset,sgm_t_scale)

      end if

c     Sea surface temperature:
      if (sst_on .eq. 1)
     >   call write_nc_var_header
     >           ('sst',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >            sst_id,1,sst_offset,sst_scale)


c     Sea surface temperature:
      if (sst_K_on .eq. 1)
     >   call write_nc_var_header
     >           ('sst_K',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >            sst_K_id,1,sst_K_offset,sst_K_scale)

c     Sea surface salinity:
      if (sss_on .eq. 1)
     >   call write_nc_var_header
     >           ('sss',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >            sss_id,1,sss_offset,sss_scale)

c     Layer thickess:
      if (nz .gt. 0 .and. leveltype .eq. 1 .and. thck_on .eq. 1)
     >   call write_nc_var_header
     >           ('thck',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >            thck_id,1,thck_offset,thck_scale)


c     Kinematics:
c     %%%%%%%%%%%

      write(*,*)
      write(*,*)'Kinematics:'

      if (nz .gt. 0) then

c        Velocity-x:
         if (u_on .eq. 1)
     >      call write_nc_var_header
     >              ('u',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               u_id,1,u_offset,u_scale)

c        Velocity-y:
         if (v_on .eq. 1)
     >      call write_nc_var_header
     >              ('v',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               v_id,1,v_offset,v_scale)

c        Volume flux-x (layers):
         if (vol_x_on .eq. 1)
     >      call write_nc_var_header
     >              ('vol_x',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               vol_x_id,1,vol_x_offset,vol_x_scale)

c        Volume flux-y (layers):
         if (vol_y_on .eq. 1)
     >      call write_nc_var_header
     >              ('vol_y',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               vol_y_id,1,vol_y_offset,vol_y_scale)

      end if


      if (nz_ice .gt. 0) then

c        Ice:
c        %%%%

         write(*,*)
         write(*,*)'Ice:'

c        Ice cover (fraction):
         if (ice_frac_on .eq. 1)
     >      call write_nc_var_header
     >              ('ice_frac',NF_SHORT,
     >               dim_options,var_n_dim,var_dim_ids,
     >               ice_frac_id,1,ice_frac_offset,ice_frac_scale)

c        Ice thickness:
         if (ice_thck_on .eq. 1)
     >      call write_nc_var_header
     >              ('ice_thck',NF_SHORT,
     >               dim_options,var_n_dim,var_dim_ids,
     >               ice_thck_id,1,ice_thck_offset,ice_thck_scale)

c        Ice velocity-x:
         if (u_ice_on .eq. 1)
     >      call write_nc_var_header
     >              ('u_ice',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               u_ice_id,1,u_ice_offset,u_ice_scale)

c        Ice velocity-y:
         if (v_ice_on .eq. 1)
     >      call write_nc_var_header
     >              ('v_ice',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >               v_ice_id,1,v_ice_offset,v_ice_scale)

c        Ice volume flux-x (layers):
         if (vol_x_ice_on .eq. 1)
     >      call write_nc_var_header
     >              ('vol_x_ice',NF_SHORT,
     >               dim_options,var_n_dim,var_dim_ids,
     >               vol_x_ice_id,1,vol_x_ice_offset,vol_x_ice_scale)

c        Ice volume flux-y (layers):
         if (vol_y_ice_on .eq. 1)
     >      call write_nc_var_header
     >              ('vol_y_ice',NF_SHORT,
     >               dim_options,var_n_dim,var_dim_ids,
     >               vol_y_ice_id,1,vol_y_ice_offset,vol_y_ice_scale)

      end if


c     Topography:
c     %%%%%%%%%%%

      write(*,*)
      write(*,*)'Topography:'

c     Sea surface height:
      if (ssh_on .eq. 1)
     >   call write_nc_var_header
     >           ('ssh',NF_SHORT,dim_options,var_n_dim,var_dim_ids,
     >            ssh_id,1,ssh_offset,ssh_scale)


c     Bottom topography:

      if (topo_on .eq. 1)
     >   call write_nc_var_header
     >           ('topo',NF_FLOAT,dim_options,var_n_dim,var_dim_ids,
     >            topo_id,1,0.,0.)


c     ***********************
c          *************

c     ***********************
c     netCDF end definitions:
c     ***********************

      nc_status = NF_ENDDEF(nc_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,53)

      write(*,*)
      write(*,*)'========================================='
      if (.not. exists) then
         write(*,*)' Done creating netCDF header on the file'
      else
         write(*,*)' Done modifying netCDF header on the file'
      end if
      write(*,*)'     ',filename(1:filesize)
      write(*,*)' The file has been set in "data mode".'
      write(*,*)'========================================='
      write(*,*)


      end
