
ccc   Subroutines on "check_nc_dims.f":
ccc    check_nc_n       - should be used to check that a dimension (and
ccc                       no. of grids in the dimension) on the netCDF
ccc                       file correspond to Fortran declarations
ccc    check_nc_xy      - should be used to check that the dimensions
ccc                       (and no. of grids in the x,y dimensions)
ccc                       on the netCDF file correspond to Fortran
ccc                       declarations
ccc    check_nc_xyz     - similar to "check_nc_xy", but for variables
ccc                       which also have a z dimension
ccc    check_nc_xyz_ice - to be included at a later stage
ccc
ccc   Functions on "check_nc_dims.f":
ccc    nc_grids_error   - checks that the no. of grids in a dimension
ccc                       on a netCDF file correspond to the associated
ccc                       declaration in the Fortran source code
ccc    nc_time_error    - checks that the length of the record dimension
ccc                       on a netCDF file is larger than or equal to
ccc                       the associated declaration in the Fortran
ccc                       source code
ccc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc


cNOTE The term "Fortranian" is used with reference to the sequence of
cNOTE  array indices, i.e.,
cNOTE   f(2,2) is stored as f(1,1), f(2,1), f(1,2), f(2,2)
cNOTE          [and NOT as  f(1,1), f(1,2), f(2,1), f(2,2)]
cNOTE  so the first dimension in "Fortranian" is f(X,) [NOT f(,X)] etc.


c     ********************
c     General description:
c     ********************

c     The subroutine "check_nc_n" should be used when reading
c     n(,t)-arrays (n is x,y,z or t) from a netCDF file, to check that
c     the dimension (and no. of grids in the dimension) on the netCDF
c     file correspond to the declaration in the Fortran source code.
c     If everything is OK, the subroutine returns with 'error'=0,
c     if non-matching information is encountered, 'error' is set to
c     ta positive number, and an error message is printed before
c     the subroutine is exited. Also, the netCDF ID of the variable
c     in question is returned as 'var_id'.


c     ***********************
c     Input/output variables:
c     ***********************

c      nc_id          (i)  - netCDF file ID
c      sphere_on      (i)  - 1: horizontal grid should be lon,lat
c                            0: horizontal grid should be x,y
c      no             (i)  - nunmber of grids should be no
c      record_on      (l)  - (.true.): there should be a record (time)
c                                       dimension
c                            (.false.): there should be no record (time)
c                                        dimension
c                                       or if the dimension to check is
c                                        the record dimension
c      exact_ndims    (i)  - 1: no. of dimensions must be exact
c                                (i.e., 2 if record dim. exists,
c                                       1 otherwise)
c                            0: no. of dimensions must be at least 1,
c                                and the n-dim must come first
c                                (in "Fortranian")
c      variable       (s)  - the name of the variable which will be
c                             checked
c      var_id         (i)  - the output variable ID for 'variable' on
c                             the file associated with 'nc_id'
c      error          (i)  - returned error status:
c                             0 - no error ('sphere_on', 'no' OK;
c                                           'record_on' OK for 'variable')
c                             1 - non-matching dimension, see printed
c                                  error message
c                             2 - type of dimension does not correspond
c                                  to 'sphere_on', see printed error
c                                  message
c                             3 - non-matching no. of grids, see printed
c                                  error message

c     NOTE:
c      If 'error' is set to a value different from 0, control will be
c      returned; no additional test(s) will be performed when 'error'
c      is set to 1 or 2..

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      l  - logical          [ (.false.)/0 or (.true.)/nonzero ]
c      s  - string           [ character*(*) ]


      subroutine check_nc_n(nc_id,sphere_on,no,record_on,exact_ndims,
     >                      variable,var_id,error)

      implicit none

      include 'netcdf.inc'

c     Input:
      integer nc_id,sphere_on,no,exact_ndims
      logical record_on
      character*(*) variable

c     Output:
      integer var_id,error

c     Local:
      integer size,var_size,nc_status,n_dim,n_id(2),n_error
      logical record_off
      character*12 dim_name

c     Functions:
      integer str_len,nc_grids_error


c     Initialize etc.:
      var_size = str_len(variable)
      if (record_on) then
         record_off = .false.
      else
         record_off = .true.
      end if


c     Find variable ID:
      nc_status = NF_INQ_VARID(nc_id, variable(1:var_size), var_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,901)


c     Check variable dimensions, no.?
      if (exact_ndims .eq. 1) then
         nc_status = NF_INQ_VARNDIMS(nc_id, var_id, n_dim)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,911)
         if (record_off .and. n_dim .ne. 1) then
            write(*,*)
            write(*,*)' Nonmatching dimensions for the variable ',
     >                variable(1:var_size)
            write(*,*)'  Declared dimensions:       ',1
            write(*,*)'  Dimensions on netCDF file: ',n_dim
            write(*,*)' ... returning from "check_nc_n".'
            write(*,*)
            error = 1
            go to 1000
         else if (record_on .and. n_dim .ne. 2) then
            write(*,*)
            write(*,*)' Nonmatching dimensions for the variable ',
     >                variable(1:var_size)
            write(*,*)'  Declared dimensions:       ',2
            write(*,*)'   (including record dimension)'
            write(*,*)'  Dimensions on netCDF file: ',n_dim
            write(*,*)' ... returning from "check_nc_n".'
            write(*,*)
            error = 1
            go to 1000
         end if
      end if


c     Check variable dimensions, which?:
      nc_status = NF_INQ_VARDIMID(nc_id, var_id, n_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,921)
      nc_status = NF_INQ_DIMNAME(nc_id, n_id(1), dim_name)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,922)
      size = str_len(dim_name)
      if (sphere_on .eq. 1) then
         if (dim_name(1:size) .ne. 'lon' .and.
     >       dim_name(1:size) .ne. 'lat') then
            write(*,*)
            write(*,*)' Unable to retrieve lon/lat results for ',
     >                variable
            write(*,*)'  (probably due to x/y grid)'
            write(*,*)' ... returning from "check_nc_n".'
            write(*,*)
            error = 2
            go to 1000
         end if
      else
         if (dim_name(1:size) .ne. 'x' .and.
     >       dim_name(1:size) .ne. 'y') then
            write(*,*)
            write(*,*)' Unable to retrieve x/y results for ',variable
            write(*,*)'  (probably due to lat/lon grid)'
            write(*,*)' ... returning from "check_nc_n".'
            write(*,*)
            error = 2
            go to 1000
         end if
      end if

c     Check dimension:
      error = nc_grids_error(nc_id, n_id(1), no, dim_name(1:size))

 1000 return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "check_nc_xy" should be used when reading
c     x,y(,t)-arrays from a netCDF file, to check that the dimensions
c     (and no. of grids in each dimension) on the netCDF file
c     correspond to the declarations of 'nx' and 'ny' in the Fortran
c     source code. If everything is OK, the subroutine returns with
c     'error'=0, if non-matching information is encountered, 'error'
c     is set to a positive number, and an error message is printed
c     before the subroutine is exited. Also, the netCDF ID of the
c     variable in question is returned as 'var_id'.


c     ***********************
c     Input/output variables:
c     ***********************

c      nc_id          (i)  - netCDF file ID
c      sphere_on      (i)  - 1: horizontal grid should be lon,lat
c                            0: horizontal grid should be x,y
c      nx, ny         (i)  - horizontal grids should be nx by ny
c      record_on      (l)  - (.true.): there should be a record (time)
c                                      dimension
c                            (.false.): there should be no record (time)
c                                       dimension
c      exact_ndims    (i)  - 1: no. of dimensions must be exact
c                                (i.e., 3 if record dim. exists,
c                                       2 otherwise)
c                            0: no. of dimensions must be at least 2,
c                                and the xy-dim.s must come first
c                                (in "Fortranian")
c      variable       (s)  - the name of the variable which will be
c                             checked
c      var_id         (i)  - the output variable ID for 'variable' on
c                             the file associated with 'nc_id'
c      error          (i)  - returned error status:
c                             0 - no error ('sphere_on', 'nx', 'ny' OK;
c                                           'record_on' OK for 'variable')
c                             1 - non-matching dimensions, see printed
c                                  error message
c                             2 - type of horizontal dimensions does not
c                                  correspond to 'sphere_on', see printed
c                                  error message
c                             3 - non-matching no. of grids in one or
c                                  both dimensions, see printed error
c                                  message

c     NOTE:
c      If 'error' is set to a value different from 0, control will be
c      returned; no additional test(s) will be performed when 'error'
c      is set to 1 or 2..

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      l  - logical          [ (.false.)/0 or (.true.)/nonzero ]
c      s  - string           [ character*(*) ]


      subroutine check_nc_xy(nc_id,sphere_on,nx,ny,record_on,
     >                       exact_ndims,variable,var_id,error)

      implicit none

      include 'netcdf.inc'

c     Input:
      integer nc_id,sphere_on,nx,ny,exact_ndims
      logical record_on
      character*(*) variable

c     Output:
      integer var_id,error

c     Local:
      integer size1,size2,var_size,nc_status,n_dim,xy_id(3),
     >        x_id,y_id,x_error,y_error
      logical record_off
      character*12 dim_name1,dim_name2

c     Functions:
      integer str_len,nc_grids_error


c     Initialize etc.:
      var_size = str_len(variable)
      if (record_on) then
         record_off = .false.
      else
         record_off = .true.
      end if


c     Find variable ID:
      nc_status = NF_INQ_VARID(nc_id, variable(1:var_size), var_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,902)


c     Check variable dimensions, no.?
      if (exact_ndims .eq. 1) then
         nc_status = NF_INQ_VARNDIMS(nc_id, var_id, n_dim)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,912)
         if (record_off .and. n_dim .ne. 2) then
            write(*,*)
            write(*,*)' Nonmatching dimensions for the variable ',
     >                variable(1:var_size)
            write(*,*)'  Declared dimensions:       ',2
            write(*,*)'  Dimensions on netCDF file: ',n_dim
            write(*,*)' ... returning from "check_nc_xy".'
            write(*,*)
            error = 1
            go to 1000
         else if (record_on .and. n_dim .ne. 3) then
            write(*,*)
            write(*,*)' Nonmatching dimensions for the variable ',
     >                variable(1:var_size)
            write(*,*)'  Declared dimensions:       ',3
            write(*,*)'   (including record dimension)'
            write(*,*)'  Dimensions on netCDF file: ',n_dim
            write(*,*)' ... returning from "check_nc_xy".'
            write(*,*)
            error = 1
            go to 1000
         end if
      end if


c     Check variable dimensions, which?:
      nc_status = NF_INQ_VARDIMID(nc_id, var_id, xy_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,923)
      nc_status = NF_INQ_DIMNAME(nc_id, xy_id(1), dim_name1)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,924)
      nc_status = NF_INQ_DIMNAME(nc_id, xy_id(2), dim_name2)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,925)
      size1 = str_len(dim_name1)
      size2 = str_len(dim_name2)
      if (sphere_on .eq. 1) then
         if ( (dim_name1(1:size1) .ne. 'lon' .and.
     >         dim_name1(1:size1) .ne. 'lat' ) .or.
     >        (dim_name2(1:size2) .ne. 'lon' .and.
     >         dim_name2(1:size2) .ne. 'lat' ) .or.
     >        (dim_name1(1:size1) .eq. dim_name2(1:size2)) ) then
            write(*,*)
            write(*,*)' Unable to retrieve lon/lat results for ',
     >                variable
            write(*,*)'  (probably due to x/y grid)'
            write(*,*)' ... returning from "check_nc_xy".'
            write(*,*)
            error = 2
            go to 1000
         end if
      else
         if ( (dim_name1(1:size1) .ne. 'x' .and.
     >         dim_name1(1:size1) .ne. 'y' ) .or.
     >        (dim_name2(1:size2) .ne. 'x' .and.
     >         dim_name2(1:size2) .ne. 'y' ) .or.
     >        (dim_name1(1:size1) .eq. dim_name2(1:size2)) ) then
            write(*,*)
            write(*,*)' Unable to retrieve x/y results for ',variable
            write(*,*)'  (probably due to lat/lon grid)'
            write(*,*)' ... returning from "check_nc_xy".'
            write(*,*)
            error = 2
            go to 1000
         end if
      end if


c     Get xy dimension IDs:
      if (sphere_on .eq. 1) then
         nc_status = NF_INQ_DIMID(nc_id, 'lon', x_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,941)
         nc_status = NF_INQ_DIMID(nc_id, 'lat', y_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,942)
      else
         nc_status = NF_INQ_DIMID(nc_id, 'x', x_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,943)
         nc_status = NF_INQ_DIMID(nc_id, 'y', y_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,944)
      end if

c     Check dimensions:
      x_error = nc_grids_error(nc_id, x_id, nx, 'x')
      y_error = nc_grids_error(nc_id, y_id, ny, 'y')

      error = max(x_error,y_error)

 1000 return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "check_nc_xyz" should be used when reading
c     x,y,z(,t)-arrays from a netCDF file, to check that the dimensions
c     (and no. of grids in each dimension) on the netCDF file
c     correspond to the declarations of 'nx', 'ny' and 'nz' in the
c     Fortran source code. If everything is OK, the subroutine returns
c     with 'error'=0, if non-matching information is encountered,
c     'error' is set to a positive number, and an error message
c     is printed before the subroutine is exited. Also, the netCDF ID
c     of the variable in question is returned as 'var_id'.


c     ***********************
c     Input/output variables:
c     ***********************

c      nc_id          (i)  - netCDF file ID
c      sphere_on      (i)  - 1: horizontal grid should be lon,lat
c                            0: horizontal grid should be x,y
c      nx, ny, nz     (i)  - grids should be nx by ny by nz
c      record_on      (l)  - (.true.): there should be a record (time)
c                                      dimension
c                            (.false.): there should be no record (time)
c                                       dimension
c      exact_ndims    (i)  - 1: no. of dimensions must be exact
c                                (i.e., 4 if record dim. exists,
c                                       3 otherwise)
c                            0: no. of dimensions must be at least 3,
c                                and the xyz-dim.s must come first
c                                (in "Fortranian")
c      variable       (s)  - the name of the variable which will be
c                             checked
c      var_id         (i)  - the output variable ID for 'variable' on
c                             the file associated with 'nc_id'
c      error          (i)  - returned error status:
c                             0 - no error ('sphere_on', 'nx', 'ny' OK;
c                                           'record_on' OK for 'variable')
c                             1 - non-matching dimensions, see printed
c                                  error message
c                             2 - type of horizontal dimensions does not
c                                  correspond to 'sphere_on', see printed
c                                  error message
c                             3 - non-matching no. of grids in one or
c                                  both dimensions, see printed error
c                                  message

c     NOTE:
c      If 'error' is set to a value different from 0, control will be
c      returned; no additional test(s) will be performed when 'error'
c      is set to 1 or 2..

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      l  - logical          [ (.false.)/0 or (.true.)/nonzero ]
c      s  - string           [ character*(*) ]


      subroutine check_nc_xyz(nc_id,sphere_on,nx,ny,nz,record_on,
     >                        exact_ndims,variable,var_id,error)

      implicit none

      include 'netcdf.inc'

c     Input:
      integer nc_id,sphere_on,nx,ny,nz,exact_ndims
      logical record_on
      character*(*) variable

c     Output:
      integer var_id,error

c     Local:
      integer size1,size2,var_size,nc_status,n_dim,xyz_id(4),
     >        x_id,y_id,z_id,x_error,y_error,z_error
      logical record_off
      character*12 dim_name1,dim_name2

c     Functions:
      integer str_len,nc_grids_error


c     Initialize etc.:
      var_size = str_len(variable)
      if (record_on) then
         record_off = .false.
      else
         record_off = .true.
      end if

c     Find variable ID:
      nc_status = NF_INQ_VARID(nc_id, variable(1:var_size), var_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,904)


c     Check variable dimensions, no.?
      if (exact_ndims .eq. 1) then
         nc_status = NF_INQ_VARNDIMS(nc_id, var_id, n_dim)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,914)
         if (record_off .and. n_dim .ne. 3) then
            write(*,*)
            write(*,*)' Nonmatching dimensions for the variable ',
     >                variable(1:var_size)
            write(*,*)'  Declared dimensions:       ',3
            write(*,*)'  Dimensions on netCDF file: ',n_dim
            write(*,*)' ... returning from "check_nc_xyz".'
            write(*,*)
            error = 1
            go to 1000
         else if (record_on .and. n_dim .ne. 4) then
            write(*,*)
            write(*,*)' Nonmatching dimensions for the variable ',
     >                variable(1:var_size)
            write(*,*)'  Declared dimensions:       ',4
            write(*,*)'   (including record dimension)'
            write(*,*)'  Dimensions on netCDF file: ',n_dim
            write(*,*)' ... returning from "check_nc_xyz".'
            write(*,*)
            error = 1
            go to 1000
         end if
      end if


c     Check variable dimensions, which?:
      nc_status = NF_INQ_VARDIMID(nc_id, var_id, xyz_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,915)
      nc_status = NF_INQ_DIMNAME(nc_id, xyz_id(1), dim_name1)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,916)
      nc_status = NF_INQ_DIMNAME(nc_id, xyz_id(2), dim_name2)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,917)
      size1 = str_len(dim_name1)
      size2 = str_len(dim_name2)
      if (sphere_on .eq. 1) then
         if ( (dim_name1(1:size1) .ne. 'lon' .and.
     >         dim_name1(1:size1) .ne. 'lat' ) .or.
     >        (dim_name2(1:size2) .ne. 'lon' .and.
     >         dim_name2(1:size2) .ne. 'lat' ) .or.
     >        (dim_name1(1:size1) .eq. dim_name2(1:size2)) ) then
            write(*,*)
            write(*,*)' Unable to retrieve lon/lat results for ',
     >                variable
            write(*,*)'  (probably due to x/y grid)'
            write(*,*)' ... returning from "check_nc_xyz".'
            write(*,*)
            error = 2
            go to 1000
         end if
      else
         if ( (dim_name1(1:size1) .ne. 'x' .and.
     >         dim_name1(1:size1) .ne. 'y' ) .or.
     >        (dim_name2(1:size2) .ne. 'x' .and.
     >         dim_name2(1:size2) .ne. 'y' ) .or.
     >        (dim_name1(1:size1) .eq. dim_name2(1:size2)) ) then
            write(*,*)
            write(*,*)' Unable to retrieve x/y results for ',variable
            write(*,*)'  (probably due to lat/lon grid)'
            write(*,*)' ... returning from "check_nc_xyz".'
            write(*,*)
            error = 2
            go to 1000
         end if
      end if


c     Get xyz dimension IDs:
      if (sphere_on .eq. 1) then
         nc_status = NF_INQ_DIMID(nc_id, 'lon', x_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,951)
         nc_status = NF_INQ_DIMID(nc_id, 'lat', y_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,952)
      else
         nc_status = NF_INQ_DIMID(nc_id, 'x', x_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,953)
         nc_status = NF_INQ_DIMID(nc_id, 'y', y_id)
         if (nc_status .ne. NF_NOERR) call handle_err(nc_status,954)
      end if
      nc_status = NF_INQ_DIMID(nc_id, 'z', z_id)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,933)

c     Check dimensions:
      x_error = nc_grids_error(nc_id, x_id, nx, 'x')
      y_error = nc_grids_error(nc_id, y_id, ny, 'y')
      z_error = nc_grids_error(nc_id, z_id, nz, 'z')

      error = max(x_error,y_error,z_error)

 1000 return
      end




c     ********************
c     General description:
c     ********************

c     The integer function "nc_grids_error" checks that the no. of
c     grids in a dimension on a netCDF file correspond to the
c     associated declaration in the Fortran source code. If this is
c     OK, the function returns the value 0, if non-matching no. of
c     grids is encountered "nc_grids_error" returns the value 3.


c     ****************
c     Input variables:
c     ****************

c      nc_id          (i)  - netCDF file ID
c      dim_id         (i)  - netCDF dimension ID for 'dim'
c      length         (i)  - should be the number of grids in 'dim'
c      dim            (s)  - the name of the dimension which will be
c                             checked

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      s  - string           [ character*(*) ]


      integer function nc_grids_error(nc_id,dim_id,length,dim)

      implicit none

      include 'netcdf.inc'

c     Input:
      integer nc_id,dim_id,length
      character*(*) dim

c     Local:
      integer nc_status,dim_length


      nc_status = NF_INQ_DIMLEN(nc_id, dim_id, dim_length)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,905)

      if (length .ne. dim_length) then
         write(*,*)
         write(*,*)' Nonmatching ',dim,' grids!'
         write(*,*)'  Declared ',dim,' grids:       ',length
         write(*,*)'  ',dim,' grids on netCDF file: ',dim_length
         write(*,*)' ... execution continues'
         write(*,*)
         nc_grids_error = 3
      else
         nc_grids_error = 0
      end if

      end



c     ********************
c     General description:
c     ********************

c     The integer function "nc_time_error" checks that the length of
c     the record dimension on a netCDF file is larger than or equal
c     to the associated declaration in the Fortran source code. If
c     this is OK, the function returns the value 0, if the declared
c     dimension is larger than the netCDF record dimension,
c     "nc_tim_error" returns the value 4.


c     ****************
c     Input variables:
c     ****************

c      nc_id          (i)  - netCDF file ID
c      dim_id         (i)  - netCDF dimension ID for 'dim'
c      length         (i)  - should be the number of grids in 'dim'
c      dim            (s)  - the name of the dimension which will be
c                             checked

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      s  - string           [ character*(*) ]

      integer function nc_time_error(nc_id,dim_id,length,dim)

      implicit none

      include 'netcdf.inc'

c     Input:
      integer nc_id,dim_id,length
      character*(*) dim

c     Local:
      integer nc_status,dim_length


      nc_status = NF_INQ_DIMLEN(nc_id, dim_id, dim_length)
      if (nc_status .ne. NF_NOERR) call handle_err(nc_status,905)

      if (length .gt. dim_length) then
         write(*,*)
         write(*,*)' Nonmatching ',dim,' dimensions for the record ID!'
         write(*,*)'  Declared ',dim,' dimension:               ',length
         write(*,*)'  ',dim,' present dimension on netCDF file: ',
     >             dim_length
         write(*,*)' ... execution continues'
         write(*,*)
         nc_time_error = 4
      else
         nc_time_error = 0
      end if

      end
