
ccc   Subroutines on "convert_int_float.f":
ccc    float2int_0D,
ccc    float2int_1D,
ccc    float2int_2D,
ccc    float2int_3D,
ccc    float2int_4D     - converts an n-dimensional value/vector/array
ccc                       from 4 byte float numbers to 4 byte integers
ccc    int2float_0D,
ccc    int2float_1D,
ccc    int2float_2D,
ccc    int2float_3D,
ccc    int2float_4D     - converts an n-dimensional value/vector/array
ccc                       from 4 byte integers to 4 byte float numbers
ccc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

cNOTE Use of these subroutines my result in errors if other missing
cNOTE values than those specified in the accompanying file
cNOTE    missing_values.inc
cNOTE are used.



c     ********************
c     General description:
c     ********************

c     The subroutine "float2int_0D" may be used for converting a real
c     value (4 byte floating point number) to an integer value, given
c     a scaling factor and an offset number:
c        integer value =
c           nearest integer [(real value - offset)/scale]

c     If the scaling factor is set to 1., and the offset value is 0.,
c     the subroutine will return the same value as the intrinsic
c     function "nint". "float2int_0D" performs the inverse procedure
c     of "int2float_0D" (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      float_number   (r)  - input floating point number
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      int_number     (i)  - output integer number (scaled and offset)

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]


      subroutine float2int_0D(float_number,offset,scale,int_number)


      implicit none

      include 'missing_values.inc'

c     Input:
      real float_number,offset,scale

c     Output:
      integer int_number

c     Local:
      real limit

      limit = 0.99*miss_val_r(1)

      if (float_number .lt. limit) then
         int_number = miss_val_i4(1)
      else
         int_number = nint( (float_number - offset)/scale )
      end if

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "float2int_1D" is the 1D (vector) version of
c     "float2int_0D", see above for a detailed description.
c     "float2int_1D" performs the inverse procedure of "int2float_1D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx             (i)  - dimension size of input and output
c                             vectors
c      float_array    (rA) - input vector of floating point numbers
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      int_array      (iA) - output vector of integer numbers
c                             (scaled and offset)

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine float2int_1D(nx,float_array,offset,scale,int_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx
      real float_array(nx),offset,scale

c     Output:
      integer int_array(nx)

c     Local:
      integer x
      real limit

      limit = 0.99*miss_val_r(1)

      do x = 1,nx
         if (float_array(x) .lt. limit) then
            int_array(x) = miss_val_i4(1)
         else
            int_array(x) = nint( (float_array(x) - offset)/scale )
         end if
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "float2int_2D" is the 2D (array) version of
c     "float2int_0D", see above for a detailed description.
c     "float2int_2D" performs the inverse procedure of "int2float_2D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx, ny         (i)  - dimension sizes of input and output
c                             arrays
c      float_array    (rA) - input vector of floating point numbers
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      int_array      (iA) - output vector of integer numbers
c                             (scaled and offset)

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine float2int_2D(nx,ny,float_array,offset,scale,
     >                          int_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx,ny
      real float_array(nx,ny),offset,scale

c     Output:
      integer int_array(nx,ny)

c     Local:
      integer x,y
      real limit

      limit = 0.99*miss_val_r(1)

      do y = 1,ny
         do x = 1,nx
            if (float_array(x,y) .lt. limit) then
               int_array(x,y) = miss_val_i4(1)
            else
               int_array(x,y) = nint( (float_array(x,y) - offset)/
     >                                scale )
            end if
         end do
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "float2int_3D" is the 3D (array) version of
c     "float2int_0D", see above for a detailed description.
c     "float2int_3D" performs the inverse procedure of "int2float_3D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx, ny, nz     (i)  - dimension sizes of input and output
c                             arrays
c      float_array    (rA) - input vector of floating point numbers
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      int_array      (iA) - output vector of integer numbers
c                             (scaled and offset)

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine float2int_3D(nx,ny,nz,float_array,offset,scale,
     >                        int_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx,ny,nz
      real float_array(nx,ny,nz),offset,scale

c     Output:
      integer int_array(nx,ny,nz)

c     Local:
      integer x,y,z
      real limit

      limit = 0.99*miss_val_r(1)

      do z = 1,nz
         do y = 1,ny
            do x = 1,nx
               if (float_array(x,y,z) .lt. limit) then
                  int_array(x,y,z) = miss_val_i4(1)
               else
                  int_array(x,y,z) = nint( (float_array(x,y,z) -
     >                                      offset)/scale )
               end if
            end do
         end do
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "float2int_4D" is the 4D (array) version of
c     "float2int_0D", see above for a detailed description.
c     "float2int_4D" performs the inverse procedure of "int2float_4D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx, ny, nz, nt (i)  - dimension sizes of input and output
c                             arrays
c      float_array    (rA) - input vector of floating point numbers
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      int_array      (iA) - output vector of integer numbers
c                             (scaled and offset)

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine float2int_4D(nx,ny,nz,nt,float_array,offset,scale,
     >                        int_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx,ny,nz,nt
      real float_array(nx,ny,nz,nt),offset,scale

c     Output:
      integer int_array(nx,ny,nz,nt)

c     Local:
      integer x,y,z,t
      real limit

      limit = 0.99*miss_val_r(1)

      do t = 1,nt
         do z = 1,nz
            do y = 1,ny
               do x = 1,nx
                  if (float_array(x,y,z,t) .lt. limit) then
                     int_array(x,y,z,t) = miss_val_i4(1)
                  else
                     int_array(x,y,z,t) = nint( (float_array(x,y,z,t) -
     >                                           offset)/scale )
               end if
               end do
            end do
         end do
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "int2float_0D" may be used for converting an
c     integer value to a real value (4 byte floating point number),
c     given a scaling factor and an offset number:
c        real value = scale * integer value + offset

c     If the scaling factor is set to 1., and the offset value is 0.,
c     the subroutine will return the same value as the intrinsic
c     function "float". "int2float_0D" performs the inverse procedure
c     of "float2int_0D" (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      int_number     (i)  - input integer number (scaled and offset)
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      float_number   (r)  - output floating point number

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]


      subroutine int2float_0D(int_number,offset,scale,float_number)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer int_number
      real offset,scale

c     Output:
      real float_number

      if (int_number .ne. miss_val_i4(1)) then
         float_number = scale*float(int_number) + offset
      else
         float_number = miss_val_r(1)
      end if

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "int2float_1D" is the 1D (vector) version of
c     "int2float_0D", see above for a detailed description.
c     "int2float_1D" performs the inverse procedure of "float2int_1D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx             (i)  - dimension size of input and output
c                             vectors
c      int_array      (iA) - input vector of integer numbers
c                             (scaled and offset)
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      float_array    (rA) - output vector of floating point numbers

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine int2float_1D(nx,int_array,offset,scale,float_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx
      integer int_array(nx)
      real offset,scale

c     Output:
      real float_array(nx)

c     Local:
      integer x

      do x = 1,nx
         if (int_array(x) .ne. miss_val_i4(1)) then
            float_array(x) = scale*float(int_array(x)) + offset
         else
            float_array(x) = miss_val_r(1)
         end if
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "int2float_2D" is the 2D (array) version of
c     "int2float_0D", see above for a detailed description.
c     "int2float_2D" performs the inverse procedure of "float2int_2D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx, ny         (i)  - dimension sizes of input and output
c                             arrays
c      int_array      (iA) - input vector of integer numbers
c                             (scaled and offset)
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      float_array    (rA) - output vector of floating point numbers

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine int2float_2D(nx,ny,int_array,offset,scale,float_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx,ny
      integer int_array(nx,ny)
      real offset,scale

c     Output:
      real float_array(nx,ny)

c     Local:
      integer x,y

      do y = 1,ny
         do x = 1,nx
            if (int_array(x,y) .ne. miss_val_i4(1)) then
               float_array(x,y) = scale*float(int_array(x,y)) + offset
            else
               float_array(x,y) = miss_val_r(1)
            end if
         end do
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "int2float_3D" is the 3D (array) version of
c     "int2float_0D", see above for a detailed description.
c     "int2float_3D" performs the inverse procedure of "float2int_3D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx, ny, nz     (i)  - dimension sizes of input and output
c                             arrays
c      int_array      (iA) - input vector of integer numbers
c                             (scaled and offset)
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      float_array    (rA) - output vector of floating point numbers

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine int2float_3D(nx,ny,nz,int_array,offset,scale,
     >                        float_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx,ny,nz
      integer int_array(nx,ny,nz)
      real offset,scale

c     Output:
      real float_array(nx,ny,nz)

c     Local:
      integer x,y,z

      do z = 1,nz
         do y = 1,ny
            do x = 1,nx
               if (int_array(x,y,z) .ne. miss_val_i4(1)) then
                  float_array(x,y,z) = scale*float(int_array(x,y,z)) +
     >                                 offset
               else
                  float_array(x,y,z) = miss_val_r(1)
               end if
            end do
         end do
      end do

      return
      end



c     ********************
c     General description:
c     ********************

c     The subroutine "int2float_4D" is the 3D (array) version of
c     "int2float_0D", see above for a detailed description.
c     "int2float_4D" performs the inverse procedure of "float2int_4D"
c     (with restrictions due to roundoff errors).


c     ***********************
c     Input/output variables:
c     ***********************

c      nx, ny, nz, nt (i)  - dimension sizes of input and output
c                             arrays
c      int_array      (iA) - input vector of integer numbers
c                             (scaled and offset)
c      offset         (r)  - offset value used n conversion
c      scale          (r)  - scaling factor used in conversion
c      float_array    (rA) - output vector of floating point numbers

c     Above:
c      i  - integer variable [ 4 byte, aka. "long" ]
c      iA - integer array    [ 4 byte, aka. "long" ]
c      r  - real variable    [ 4 byte, aka. "float" ]
c      rA - real array       [ 4 byte, aka. "float" ]


      subroutine int2float_4D(nx,ny,nz,nt,int_array,offset,scale,
     >                        float_array)


      implicit none

      include 'missing_values.inc'

c     Input:
      integer nx,ny,nz,nt
      integer int_array(nx,ny,nz,nt)
      real offset,scale

c     Output:
      real float_array(nx,ny,nz,nt)

c     Local:
      integer x,y,z,t

      do t = 1,nt
         do z = 1,nz
            do y = 1,ny
               do x = 1,nx
                  if (int_array(x,y,z,t) .ne. miss_val_i4(1)) then
                     float_array(x,y,z,t) =
     >                    scale*float(int_array(x,y,z,t)) + offset
                  else
                     float_array(x,y,z,t) = miss_val_r(1)
                  end if
               end do
            end do
         end do
      end do

      return
      end

